@@ -17,12 +17,30 @@
 
 #include "private.h"
 
+#if LUV_UV_VERSION_GEQ(1, 28, 0)
+typedef struct {
+  uv_dir_t* handle;
+  int dirents_ref; /* handle has been closed if this is LUA_NOREF */
+} luv_dir_t;
+#endif
+
 static uv_fs_t* luv_check_fs(lua_State* L, int index) {
+  if (luaL_testudata(L, index, "uv_fs") != NULL) {
+    return (uv_fs_t*)lua_touserdata(L, index);
+  }
   uv_fs_t* req = (uv_fs_t*)luaL_checkudata(L, index, "uv_req");
   luaL_argcheck(L, req->type == UV_FS && req->data, index, "Expected uv_fs_t");
   return req;
 }
 
+static int luv_fs_gc(lua_State* L) {
+  uv_fs_t* req = luv_check_fs(L, 1);
+  luv_cleanup_req(L, (luv_req_t*)req->data);
+  req->data = NULL;
+  uv_fs_req_cleanup(req);
+  return 0;
+}
+
 static void luv_push_timespec_table(lua_State* L, const uv_timespec_t* t) {
   lua_createtable(L, 0, 2);
   lua_pushinteger(L, t->tv_sec);
@@ -333,14 +351,15 @@ static int push_fs_result(lua_State* L, uv_fs_t* req) {
       luaL_unref(L, LUA_REGISTRYINDEX, data->data_ref);
       data->data_ref = LUA_NOREF;
 
-      (*(void**)lua_newuserdata(L, sizeof(void*))) = dir;
-      lua_pushfstring(L, "uv_dir:%p", dir);
-      dir->dirents = lua_newuserdata(L, sizeof(uv_dirent_t)*nentries);
-      dir->nentries = nentries;
-      lua_rawset(L, LUA_REGISTRYINDEX);
+      luv_dir_t* luv_dir = lua_newuserdata(L, sizeof(*luv_dir));
       luaL_getmetatable(L, "uv_dir");
       lua_setmetatable(L, -2);
 
+      luv_dir->handle = dir;
+      luv_dir->handle->dirents = lua_newuserdata(L, sizeof(uv_dirent_t)*nentries);
+      luv_dir->dirents_ref = luaL_ref(L, LUA_REGISTRYINDEX);
+      luv_dir->handle->nentries = nentries;
+
       return 1;
     }
     case UV_FS_READDIR: {
@@ -372,6 +391,11 @@ static int push_fs_result(lua_State* L, uv_fs_t* req) {
 
 static void luv_fs_cb(uv_fs_t* req) {
   luv_req_t* data = (luv_req_t*)req->data;
+  // This can happen if luv_fs_cb is called during loop gc. For example, this can happen
+  // when the async version of fs_scandir is called but the loop is never run before the process exits.
+  //
+  // TODO: A more comprehensive fix for this problem would be related to https://github.com/luvit/luv/issues/437
+  if (data == NULL) return;
   lua_State* L = data->ctx->L;
 
   int nargs = push_fs_result(L, req);
@@ -404,14 +428,14 @@ static void luv_fs_cb(uv_fs_t* req) {
 // variable 'nargs' to the number of return values
 #define FS_CALL_NORETURN(func, req, ...) {                \
   int ret, sync;                                          \
-  luv_req_t* data = (luv_req_t*)req->data;                \
-  sync = data->callback_ref == LUA_NOREF;                 \
-  ret = uv_fs_##func(data->ctx->loop, req, __VA_ARGS__,   \
+  luv_req_t* lreq = (luv_req_t*)req->data;                \
+  sync = lreq->callback_ref == LUA_NOREF;                 \
+  ret = func(lreq->ctx->loop, req, __VA_ARGS__,           \
                      sync ? NULL : luv_fs_cb);            \
   if (req->fs_type != UV_FS_ACCESS && ret < 0) {          \
     lua_pushnil(L);                                       \
     if (fs_req_has_dest_path(req)) {                      \
-      lua_rawgeti(L, LUA_REGISTRYINDEX, data->data_ref);  \
+      lua_rawgeti(L, LUA_REGISTRYINDEX, lreq->data_ref);  \
       const char* dest_path = lua_tostring(L, -1);        \
       lua_pop(L, 1);                                      \
       lua_pushfstring(L, "%s: %s: %s -> %s",              \
@@ -430,21 +454,23 @@ static void luv_fs_cb(uv_fs_t* req) {
           uv_strerror(req->result));                      \
     }                                                     \
     lua_pushstring(L, uv_err_name(req->result));          \
-    luv_cleanup_req(L, data);                             \
-    req->data = NULL;                                     \
-    uv_fs_req_cleanup(req);                               \
+    if(req->fs_type != UV_FS_SCANDIR) {                   \
+      luv_cleanup_req(L, lreq);                           \
+      req->data = NULL;                                   \
+      uv_fs_req_cleanup(req);                             \
+    }                                                     \
     nargs = 3;                                            \
   }                                                       \
   else if (sync) {                                        \
     nargs = push_fs_result(L, req);                       \
     if(req->fs_type != UV_FS_SCANDIR) {                   \
-      luv_cleanup_req(L, data);                           \
+      luv_cleanup_req(L, lreq);                           \
       req->data = NULL;                                   \
       uv_fs_req_cleanup(req);                             \
     }                                                     \
   }                                                       \
   else {                                                  \
-    lua_rawgeti(L, LUA_REGISTRYINDEX, data->req_ref);     \
+    lua_rawgeti(L, LUA_REGISTRYINDEX, lreq->req_ref);     \
     nargs = 1;                                            \
   }                                                       \
 }
@@ -461,7 +487,7 @@ static int luv_fs_close(lua_State* L) {
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(close, req, file);
+  FS_CALL(uv_fs_close, req, file);
 }
 
 static int luv_fs_open(lua_State* L) {
@@ -472,7 +498,7 @@ static int luv_fs_open(lua_State* L) {
   int ref = luv_check_continuation(L, 4);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(open, req, path, flags, mode);
+  FS_CALL(uv_fs_open, req, path, flags, mode);
 }
 
 static int luv_fs_read(lua_State* L) {
@@ -482,6 +508,7 @@ static int luv_fs_read(lua_State* L) {
   // -1 offset means "the current file offset is used and updated"
   int64_t offset = -1;
   int ref;
+  char* data;
   // both offset and callback are optional
   if (luv_is_callable(L, 3) && lua_isnoneornil(L, 4)) {
     ref = luv_check_continuation(L, 3);
@@ -490,7 +517,7 @@ static int luv_fs_read(lua_State* L) {
     offset = luaL_optinteger(L, 3, offset);
     ref = luv_check_continuation(L, 4);
   }
-  char* data = (char*)malloc(len);
+  data = (char*)malloc(len);
   if (!data) {
     luaL_unref(L, LUA_REGISTRYINDEX, ref);
     return luaL_error(L, "Failure to allocate buffer");
@@ -500,7 +527,7 @@ static int luv_fs_read(lua_State* L) {
   req->data = luv_setup_req(L, ctx, ref);
   // TODO: find out why we can't just use req->ptr for the base
   ((luv_req_t*)req->data)->data = buf.base;
-  FS_CALL(read, req, file, &buf, 1, offset);
+  FS_CALL(uv_fs_read, req, file, &buf, 1, offset);
 }
 
 static int luv_fs_unlink(lua_State* L) {
@@ -509,7 +536,7 @@ static int luv_fs_unlink(lua_State* L) {
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(unlink, req, path);
+  FS_CALL(uv_fs_unlink, req, path);
 }
 
 static int luv_fs_write(lua_State* L) {
@@ -531,7 +558,7 @@ static int luv_fs_write(lua_State* L) {
   size_t count;
   uv_buf_t* bufs = luv_check_bufs(L, 2, &count, (luv_req_t*)req->data);
   int nargs;
-  FS_CALL_NORETURN(write, req, file, bufs, count, offset);
+  FS_CALL_NORETURN(uv_fs_write, req, file, bufs, count, offset);
   free(bufs);
   return nargs;
 }
@@ -543,7 +570,7 @@ static int luv_fs_mkdir(lua_State* L) {
   int ref = luv_check_continuation(L, 3);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(mkdir, req, path, mode);
+  FS_CALL(uv_fs_mkdir, req, path, mode);
 }
 
 static int luv_fs_mkdtemp(lua_State* L) {
@@ -552,7 +579,7 @@ static int luv_fs_mkdtemp(lua_State* L) {
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(mkdtemp, req, tpl);
+  FS_CALL(uv_fs_mkdtemp, req, tpl);
 }
 
 #if LUV_UV_VERSION_GEQ(1, 34, 0)
@@ -562,7 +589,7 @@ static int luv_fs_mkstemp(lua_State* L) {
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(mkstemp, req, tpl);
+  FS_CALL(uv_fs_mkstemp, req, tpl);
 }
 #endif
 
@@ -572,7 +599,7 @@ static int luv_fs_rmdir(lua_State* L) {
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(rmdir, req, path);
+  FS_CALL(uv_fs_rmdir, req, path);
 }
 
 static int luv_fs_scandir(lua_State* L) {
@@ -581,20 +608,15 @@ static int luv_fs_scandir(lua_State* L) {
   int flags = 0; // TODO: find out what these flags are.
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
-  req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(scandir, req, path, flags);
+  req->data = luv_setup_req_with_mt(L, ctx, ref, "uv_fs");
+  FS_CALL(uv_fs_scandir, req, path, flags);
 }
 
 static int luv_fs_scandir_next(lua_State* L) {
   uv_fs_t* req = luv_check_fs(L, 1);
   uv_dirent_t ent;
   int ret = uv_fs_scandir_next(req, &ent);
-  if (ret == UV_EOF) {
-    luv_cleanup_req(L, (luv_req_t*)req->data);
-    req->data = NULL;
-    uv_fs_req_cleanup(req);
-    return 0;
-  }
+  if (ret == UV_EOF) return 0;
   if (ret < 0) return luv_error(L, ret);
   return luv_push_dirent(L, &ent, 0);
 }
@@ -605,7 +627,7 @@ static int luv_fs_stat(lua_State* L) {
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(stat, req, path);
+  FS_CALL(uv_fs_stat, req, path);
 }
 
 static int luv_fs_fstat(lua_State* L) {
@@ -614,7 +636,7 @@ static int luv_fs_fstat(lua_State* L) {
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(fstat, req, file);
+  FS_CALL(uv_fs_fstat, req, file);
 }
 
 static int luv_fs_lstat(lua_State* L) {
@@ -623,7 +645,7 @@ static int luv_fs_lstat(lua_State* L) {
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(lstat, req, path);
+  FS_CALL(uv_fs_lstat, req, path);
 }
 
 static int luv_fs_rename(lua_State* L) {
@@ -636,7 +658,7 @@ static int luv_fs_rename(lua_State* L) {
   // ref the dest path so that we can print it in the error message
   lua_pushvalue(L, 2);
   ((luv_req_t*)req->data)->data_ref = luaL_ref(L, LUA_REGISTRYINDEX);
-  FS_CALL(rename, req, path, new_path);
+  FS_CALL(uv_fs_rename, req, path, new_path);
 }
 
 static int luv_fs_fsync(lua_State* L) {
@@ -645,7 +667,7 @@ static int luv_fs_fsync(lua_State* L) {
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(fsync, req, file);
+  FS_CALL(uv_fs_fsync, req, file);
 }
 
 static int luv_fs_fdatasync(lua_State* L) {
@@ -654,7 +676,7 @@ static int luv_fs_fdatasync(lua_State* L) {
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(fdatasync, req, file);
+  FS_CALL(uv_fs_fdatasync, req, file);
 }
 
 static int luv_fs_ftruncate(lua_State* L) {
@@ -664,7 +686,7 @@ static int luv_fs_ftruncate(lua_State* L) {
   int ref = luv_check_continuation(L, 3);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(ftruncate, req, file, offset);
+  FS_CALL(uv_fs_ftruncate, req, file, offset);
 }
 
 static int luv_fs_sendfile(lua_State* L) {
@@ -676,7 +698,7 @@ static int luv_fs_sendfile(lua_State* L) {
   int ref = luv_check_continuation(L, 5);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(sendfile, req, out_fd, in_fd, in_offset, length);
+  FS_CALL(uv_fs_sendfile, req, out_fd, in_fd, in_offset, length);
 }
 
 static int luv_fs_access(lua_State* L) {
@@ -686,7 +708,7 @@ static int luv_fs_access(lua_State* L) {
   int ref = luv_check_continuation(L, 3);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(access, req, path, amode);
+  FS_CALL(uv_fs_access, req, path, amode);
 }
 
 static int luv_fs_chmod(lua_State* L) {
@@ -696,7 +718,7 @@ static int luv_fs_chmod(lua_State* L) {
   int ref = luv_check_continuation(L, 3);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(chmod, req, path, mode);
+  FS_CALL(uv_fs_chmod, req, path, mode);
 }
 
 static int luv_fs_fchmod(lua_State* L) {
@@ -706,7 +728,7 @@ static int luv_fs_fchmod(lua_State* L) {
   int ref = luv_check_continuation(L, 3);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(fchmod, req, file, mode);
+  FS_CALL(uv_fs_fchmod, req, file, mode);
 }
 
 static int luv_fs_utime(lua_State* L) {
@@ -717,7 +739,7 @@ static int luv_fs_utime(lua_State* L) {
   int ref = luv_check_continuation(L, 4);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(utime, req, path, atime, mtime);
+  FS_CALL(uv_fs_utime, req, path, atime, mtime);
 }
 
 static int luv_fs_futime(lua_State* L) {
@@ -728,7 +750,7 @@ static int luv_fs_futime(lua_State* L) {
   int ref = luv_check_continuation(L, 4);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(futime, req, file, atime, mtime);
+  FS_CALL(uv_fs_futime, req, file, atime, mtime);
 }
 
 #if LUV_UV_VERSION_GEQ(1, 36, 0)
@@ -740,7 +762,7 @@ static int luv_fs_lutime(lua_State* L) {
   int ref = luv_check_continuation(L, 4);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(lutime, req, path, atime, mtime);
+  FS_CALL(uv_fs_lutime, req, path, atime, mtime);
 }
 #endif
 
@@ -754,7 +776,7 @@ static int luv_fs_link(lua_State* L) {
   // ref the dest path so that we can print it in the error message
   lua_pushvalue(L, 2);
   ((luv_req_t*)req->data)->data_ref = luaL_ref(L, LUA_REGISTRYINDEX);
-  FS_CALL(link, req, path, new_path);
+  FS_CALL(uv_fs_link, req, path, new_path);
 }
 
 static int luv_fs_symlink(lua_State* L) {
@@ -788,7 +810,7 @@ static int luv_fs_symlink(lua_State* L) {
   // ref the dest path so that we can print it in the error message
   lua_pushvalue(L, 2);
   ((luv_req_t*)req->data)->data_ref = luaL_ref(L, LUA_REGISTRYINDEX);
-  FS_CALL(symlink, req, path, new_path, flags);
+  FS_CALL(uv_fs_symlink, req, path, new_path, flags);
 }
 
 static int luv_fs_readlink(lua_State* L) {
@@ -797,7 +819,7 @@ static int luv_fs_readlink(lua_State* L) {
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(readlink, req, path);
+  FS_CALL(uv_fs_readlink, req, path);
 }
 
 #if LUV_UV_VERSION_GEQ(1, 8, 0)
@@ -807,7 +829,7 @@ static int luv_fs_realpath(lua_State* L) {
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(realpath, req, path);
+  FS_CALL(uv_fs_realpath, req, path);
 }
 #endif
 
@@ -819,7 +841,7 @@ static int luv_fs_chown(lua_State* L) {
   int ref = luv_check_continuation(L, 4);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(chown, req, path, uid, gid);
+  FS_CALL(uv_fs_chown, req, path, uid, gid);
 }
 
 static int luv_fs_fchown(lua_State* L) {
@@ -830,7 +852,7 @@ static int luv_fs_fchown(lua_State* L) {
   int ref = luv_check_continuation(L, 4);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(fchown, req, file, uid, gid);
+  FS_CALL(uv_fs_fchown, req, file, uid, gid);
 }
 
 #if LUV_UV_VERSION_GEQ(1, 21, 0)
@@ -842,7 +864,7 @@ static int luv_fs_lchown(lua_State* L) {
   int ref = luv_check_continuation(L, 4);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(lchown, req, path, uid, gid);
+  FS_CALL(uv_fs_lchown, req, path, uid, gid);
 }
 #endif
 
@@ -883,13 +905,13 @@ static int luv_fs_copyfile(lua_State*L) {
   // ref the dest path so that we can print it in the error message
   lua_pushvalue(L, 2);
   ((luv_req_t*)req->data)->data_ref = luaL_ref(L, LUA_REGISTRYINDEX);
-  FS_CALL(copyfile, req, path, new_path, flags);
+  FS_CALL(uv_fs_copyfile, req, path, new_path, flags);
 }
 #endif
 
 #if LUV_UV_VERSION_GEQ(1, 28, 0)
-static uv_dir_t* luv_check_dir(lua_State* L, int idx) {
-  uv_dir_t* dir = *(uv_dir_t**)luaL_checkudata(L, idx, "uv_dir");
+static luv_dir_t* luv_check_dir(lua_State* L, int idx) {
+  luv_dir_t* dir = (luv_dir_t*)luaL_checkudata(L, idx, "uv_dir");
   return dir;
 }
 
@@ -905,51 +927,50 @@ static int luv_fs_opendir(lua_State* L) {
   lua_pushinteger(L, nentries);
   ((luv_req_t*)req->data)->data_ref = luaL_ref(L, LUA_REGISTRYINDEX);
 
-  FS_CALL(opendir, req, path);
+  FS_CALL(uv_fs_opendir, req, path);
 }
 
 static int luv_fs_readdir(lua_State* L) {
   luv_ctx_t* ctx = luv_context(L);
   uv_fs_t *req;
-  uv_dir_t* dir = luv_check_dir(L, 1);
+  luv_dir_t* dir = luv_check_dir(L, 1);
   int ref = luv_check_continuation(L, 2);
 
   req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(readdir, req, dir);
+  FS_CALL(uv_fs_readdir, req, dir->handle);
 }
 
 static int luv_fs_closedir(lua_State* L) {
   luv_ctx_t* ctx = luv_context(L);
-  uv_dir_t* dir = luv_check_dir(L, 1);
+  luv_dir_t* dir = luv_check_dir(L, 1);
   int ref = luv_check_continuation(L, 2);
+
+  luaL_unref(L, LUA_REGISTRYINDEX, dir->dirents_ref);
+  dir->dirents_ref = LUA_NOREF;
+
   uv_fs_t *req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  lua_pushfstring(L, "uv_dir:%p", dir);
-  lua_pushnil(L);
-  lua_rawset(L, LUA_REGISTRYINDEX);
-  FS_CALL(closedir, req, dir);
+  FS_CALL(uv_fs_closedir, req, dir->handle);
 }
 
 static int luv_fs_dir_tostring(lua_State* L) {
-  uv_dir_t* dir = luv_check_dir(L, 1);
+  luv_dir_t* dir = luv_check_dir(L, 1);
   lua_pushfstring(L, "uv_dir_t: %p", dir);
   return 1;
 }
 
 static int luv_fs_dir_gc(lua_State* L) {
-  uv_dir_t* dir = luv_check_dir(L, 1);
-  lua_pushfstring(L, "uv_dir:%p", dir);
-  lua_rawget(L, LUA_REGISTRYINDEX);
-  if (!lua_isnil(L, -1)) {
+  luv_dir_t* dir = luv_check_dir(L, 1);
+  if (dir->dirents_ref != LUA_NOREF) {
     uv_fs_t req;
     luv_ctx_t* ctx = luv_context(L);
 
-    uv_fs_closedir(ctx->loop, &req, dir, NULL);
+    luaL_unref(L, LUA_REGISTRYINDEX, dir->dirents_ref);
+    dir->dirents_ref = LUA_NOREF;
+
+    uv_fs_closedir(ctx->loop, &req, dir->handle, NULL);
     uv_fs_req_cleanup(&req);
-    lua_pushfstring(L, "uv_dir:%p", dir);
-    lua_pushnil(L);
-    lua_rawset(L, LUA_REGISTRYINDEX);
   }
   lua_pop(L, 1);
 
@@ -964,7 +985,7 @@ static int luv_fs_statfs(lua_State* L) {
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(statfs, req, path);
+  FS_CALL(uv_fs_statfs, req, path);
 }
 #endif
 
