@@ -305,6 +305,9 @@ static const luaL_Reg luv_functions[] = {
   {"os_homedir", luv_os_homedir},
   {"os_tmpdir", luv_os_tmpdir},
   {"os_get_passwd", luv_os_get_passwd},
+#endif
+#if LUV_UV_VERSION_GEQ(1, 44, 0)
+  {"available_parallelism", luv_available_parallelism},
 #endif
   {"cpu_info", luv_cpu_info},
   {"cwd", luv_cwd},
@@ -369,12 +372,22 @@ static const luaL_Reg luv_functions[] = {
   {"random", luv_random},
 #endif
   {"sleep", luv_sleep},
+#if LUV_UV_VERSION_GEQ(1, 45, 0)
+  {"cpumask_size", luv_cpumask_size},
+  {"get_available_memory", luv_get_available_memory},
+  {"clock_gettime", luv_clock_gettime},
+#endif
 
   // thread.c
   {"new_thread", luv_new_thread},
   {"thread_equal", luv_thread_equal},
   {"thread_self", luv_thread_self},
   {"thread_join", luv_thread_join},
+#if LUV_UV_VERSION_GEQ(1, 45, 0)
+  {"thread_getaffinity", luv_thread_getaffinity},
+  {"thread_setaffinity", luv_thread_setaffinity},
+  {"thread_getcpu", luv_thread_getcpu},
+#endif
 
   // work.c
   {"new_work", luv_new_work},
@@ -389,6 +402,9 @@ static const luaL_Reg luv_functions[] = {
 #if LUV_UV_VERSION_GEQ(1, 39, 0)
   {"metrics_idle_time", luv_metrics_idle_time},
 #endif
+#if LUV_UV_VERSION_GEQ(1, 45, 0)
+  {"metrics_info", luv_metrics_info},
+#endif
 
   {NULL, NULL}
 };
@@ -645,6 +661,17 @@ static void luv_req_init(lua_State* L) {
   luaL_newlib(L, luv_req_methods);
   lua_setfield(L, -2, "__index");
   lua_pop(L, 1);
+
+  // Only used for things that need to be garbage collected
+  // (e.g. the req when using uv_fs_scandir)
+  luaL_newmetatable(L, "uv_fs");
+  lua_pushcfunction(L, luv_req_tostring);
+  lua_setfield(L, -2, "__tostring");
+  luaL_newlib(L, luv_req_methods);
+  lua_setfield(L, -2, "__index");
+  lua_pushcfunction(L, luv_fs_gc);
+  lua_setfield(L, -2, "__gc");
+  lua_pop(L, 1);
 }
 
 // Call lua function, will pop nargs values from top of vm stack and push some
@@ -853,5 +880,17 @@ LUALIB_API int luaopen_luv (lua_State* L) {
 
   luv_constants(L);
   lua_setfield(L, -2, "constants");
+
+#define XX(code, _)                 \
+  lua_pushliteral(L, #code);        \
+  lua_pushinteger(L, UV_ ## code);  \
+  lua_rawset(L, -3);
+
+  lua_newtable(L);
+  UV_ERRNO_MAP(XX)
+  lua_setfield(L, -2, "errno");
+
+#undef XX
+
   return 1;
 }
