@@ -22,9 +22,14 @@ typedef struct {
   int len;
   int argc;
   luv_thread_arg_t args;
+
+  // private fields, avoid thread be released before it done
+  lua_State *L;
+  int ref;
+  uv_async_t notify;
 } luv_thread_t;
 
-static lua_State* luv_thread_acquire_vm() {
+static lua_State* luv_thread_acquire_vm(void) {
   lua_State* L = luaL_newstate();
 
   // Add in the lua standard libraries
@@ -65,7 +70,7 @@ static int luv_thread_arg_set(lua_State* L, luv_thread_arg_t* args, int idx, int
   idx = idx > 0 ? idx : 1;
   i = idx;
   args->flags = flags;
-  while (i <= top && i <= LUV_THREAD_MAXNUM_ARG + idx)
+  while (i <= top && i < LUV_THREAD_MAXNUM_ARG + idx)
   {
     luv_val_t *arg = args->argv + i - idx;
     arg->type = lua_type(L, i);
@@ -103,11 +108,10 @@ static int luv_thread_arg_set(lua_State* L, luv_thread_arg_t* args, int idx, int
       }
       break;
     default:
-      fprintf(stderr, "Error: thread arg not support type '%s' at %d",
-        lua_typename(L, arg->type), i);
-      arg->val.str.base = NULL;
-      arg->val.str.len = 0;
-      break;
+      args->argc = i - idx;
+      lua_pushinteger(L, arg->type);
+      lua_pushinteger(L, i - idx + 1);
+      return -1;
     }
     i++;
   }
@@ -207,9 +211,19 @@ static int luv_thread_arg_push(lua_State* L, luv_thread_arg_t* args, int flags)
   return i;
 }
 
-int thread_dump(lua_State* L, const void* p, size_t sz, void* B) {
+static int luv_thread_arg_error(lua_State *L) {
+  int type = lua_tointeger(L, -2);
+  int pos = lua_tointeger(L, -1);
+  lua_pop(L, 2);
+  return luaL_error(L, "Error: thread arg not support type '%s' at %d",
+    lua_typename(L, type), pos);
+}
+
+static int thread_dump (lua_State *L, const void *b, size_t size, void *ud) {
+  luaL_Buffer *B = (luaL_Buffer *)ud;
   (void)L;
-  luaL_addlstring((luaL_Buffer*) B, (const char*) p, sz);
+
+  luaL_addlstring(B, (const char *)b, size);
   return 0;
 }
 
@@ -217,15 +231,19 @@ static int luv_thread_dumped(lua_State* L, int idx) {
   if (lua_isstring(L, idx)) {
     lua_pushvalue(L, idx);
   } else {
-    int ret;
-    luaL_Buffer b;
+    int ret, top;
+    luaL_Buffer B;
+
+    // In Lua >= 5.4.3, luaL_buffinit pushes a value onto the stack, so it needs to be called
+    // here to ensure that the function is at the top of the stack during the lua_dump call
+    luaL_buffinit(L, &B);
     luaL_checktype(L, idx, LUA_TFUNCTION);
     lua_pushvalue(L, idx);
-    luaL_buffinit(L, &b);
-    ret = lua_dump(L, thread_dump, &b, 1);
-    lua_pop(L, 1);
+    top = lua_gettop(L);
+    ret = lua_dump(L, thread_dump, &B, 1);
+    lua_remove(L, top);
     if (ret==0) {
-      luaL_pushresult(&b);
+      luaL_pushresult(&B);
     } else
       luaL_error(L, "Error: unable to dump given function");
   }
@@ -240,8 +258,6 @@ static luv_thread_t* luv_check_thread(lua_State* L, int index) {
 static int luv_thread_gc(lua_State* L) {
   luv_thread_t* tid = luv_check_thread(L, 1);
   free(tid->code);
-  tid->code = NULL;
-  tid->len = 0;
   luv_thread_arg_clear(L, &tid->args, LUVF_THREAD_SIDE_MAIN);
   return 0;
 }
@@ -257,13 +273,17 @@ static void luv_thread_cb(void* varg) {
   //acquire vm and get top
   luv_thread_t* thd = (luv_thread_t*)varg;
   lua_State* L = acquire_vm_cb();
+  luv_ctx_t *ctx = luv_context(L);
+
+  lua_pushboolean(L, 1);
+  lua_setglobal(L, "_THREAD");
 
   //push lua function, thread entry
   if (luaL_loadbuffer(L, thd->code, thd->len, "=thread") == 0) {
     //push parameter for real thread function
     int i = luv_thread_arg_push(L, &thd->args, LUVF_THREAD_SIDE_CHILD);
 
-    luv_cfpcall(L, i, 0, 0);
+    ctx->thrd_pcall(L, i, 0, 0);
     luv_thread_arg_clear(L, &thd->args, LUVF_THREAD_SIDE_CHILD);
   } else {
     fprintf(stderr, "Uncaught Error in thread: %s\n", lua_tostring(L, -1));
@@ -271,15 +291,32 @@ static void luv_thread_cb(void* varg) {
     lua_pop(L, 1);
   }
 
+  uv_async_send(&thd->notify);
   release_vm_cb(L);
 }
 
+static void luv_thread_notify_close_cb(uv_handle_t *handle) {
+  luv_thread_t *thread = handle->data;
+  if (thread->handle != 0)
+    uv_thread_join(&thread->handle);
+
+  luaL_unref(thread->L, LUA_REGISTRYINDEX, thread->ref);
+  thread->ref = LUA_NOREF;
+  thread->L = NULL;
+}
+
+static void luv_thread_exit_cb(uv_async_t* handle) {
+  uv_close((uv_handle_t*)handle, luv_thread_notify_close_cb);
+}
+
 static int luv_new_thread(lua_State* L) {
   int ret;
   size_t len;
   char* code;
   luv_thread_t* thread;
   int cbidx = 1;
+  luv_ctx_t* ctx = luv_context(L);
+
 #if LUV_UV_VERSION_GEQ(1, 26, 0)
   uv_thread_options_t options;
   options.flags = UV_THREAD_NO_FLAGS;
@@ -317,22 +354,120 @@ static int luv_new_thread(lua_State* L) {
   lua_remove(L, -2);
   //clear in luv_thread_gc or in child threads
   thread->argc = luv_thread_arg_set(L, &thread->args, cbidx+1, lua_gettop(L) - 1, LUVF_THREAD_SIDE_MAIN);
+  if (thread->argc < 0) {
+    return luv_thread_arg_error(L);
+  }
   thread->len = len;
 
+  thread->notify.data = thread;
+  thread->ref = LUA_NOREF;
+  thread->L = L;
+  ret = uv_async_init(ctx->loop, &thread->notify, luv_thread_exit_cb);
+  if (ret < 0)
+    return luv_error(L, ret);
+
+  lua_pushvalue(L, -1);
+  thread->ref = luaL_ref(L, LUA_REGISTRYINDEX);
+
 #if LUV_UV_VERSION_GEQ(1, 26, 0)
   ret = uv_thread_create_ex(&thread->handle, &options, luv_thread_cb, thread);
 #else
   ret = uv_thread_create(&thread->handle, luv_thread_cb, thread);
 #endif
-  if (ret < 0) return luv_error(L, ret);
+  if (ret < 0) {
+    uv_close((uv_handle_t*)&thread->notify, luv_thread_notify_close_cb);
+    return luv_error(L, ret);
+  }
 
   return 1;
 }
 
+#if LUV_UV_VERSION_GEQ(1, 45, 0)
+static int luv_thread_getaffinity(lua_State* L) {
+  luv_thread_t* tid = luv_check_thread(L, 1);
+  int default_mask_size = uv_cpumask_size();
+  if (default_mask_size < 0) {
+    return luv_error(L, default_mask_size);
+  }
+  int mask_size = luaL_optinteger(L, 2, default_mask_size);
+  if (mask_size < default_mask_size) {
+    return luaL_argerror(L, 2, lua_pushfstring(L, "cpumask size must be >= %d (from cpumask_size()), got %d", default_mask_size, mask_size));
+  }
+  char* cpumask = malloc(mask_size);
+  int ret = uv_thread_getaffinity(&tid->handle, cpumask, mask_size);
+  if (ret < 0) {
+    free(cpumask);
+    return luv_error(L, ret);
+  }
+  lua_newtable(L);
+  for (int i = 0; i < mask_size; i++) {
+    lua_pushboolean(L, cpumask[i]);
+    lua_rawseti(L, -2, i + 1);
+  }
+  free(cpumask);
+  return 1;
+}
+
+static int luv_thread_setaffinity(lua_State* L) {
+  luv_thread_t* tid = luv_check_thread(L, 1);
+  luaL_checktype(L, 2, LUA_TTABLE);
+  int get_old_mask = lua_toboolean(L, 3);
+  int min_mask_size = uv_cpumask_size();
+  if (min_mask_size < 0) {
+    return luv_error(L, min_mask_size);
+  }
+  int mask_size = lua_rawlen(L, 2);
+  // If the provided table's length is not at least min_mask_size,
+  // we'll use the min_mask_size and fill in any missing values with
+  // false.
+  if (mask_size < min_mask_size) {
+    mask_size = min_mask_size;
+  }
+  char* cpumask = malloc(mask_size);
+  for (int i = 0; i < mask_size; i++) {
+    lua_rawgeti(L, 2, i+1);
+    int val = lua_toboolean(L, -1);
+    cpumask[i] = val;
+    lua_pop(L, 1);
+  }
+  char* oldmask = get_old_mask ? malloc(mask_size) : NULL;
+  int ret = uv_thread_setaffinity(&tid->handle, cpumask, oldmask, mask_size);
+  // Done with cpumask at this point
+  free(cpumask);
+  if (ret < 0) {
+    if (get_old_mask) free(oldmask);
+    return luv_error(L, ret);
+  }
+  if (get_old_mask) {
+    lua_newtable(L);
+    for (int i = 0; i < mask_size; i++) {
+      lua_pushboolean(L, oldmask[i]);
+      lua_rawseti(L, -2, i + 1);
+    }
+    free(oldmask);
+  }
+  else {
+    lua_pushboolean(L, 1);
+  }
+  return 1;
+}
+
+static int luv_thread_getcpu(lua_State* L) {
+  int ret = uv_thread_getcpu();
+  if (ret < 0) return luv_error(L, ret);
+  // Make the returned value start at 1 to match how getaffinity/setaffinity
+  // masks are implemented (they use array-like tables so the first
+  // CPU is index 1).
+  lua_pushinteger(L, ret + 1);
+  return 1;
+}
+#endif
+
 static int luv_thread_join(lua_State* L) {
   luv_thread_t* tid = luv_check_thread(L, 1);
   int ret = uv_thread_join(&tid->handle);
   if (ret < 0) return luv_error(L, ret);
+  tid->handle = 0;
   lua_pushboolean(L, 1);
   return 1;
 }
@@ -360,6 +495,10 @@ static int luv_thread_equal(lua_State* L) {
 static const luaL_Reg luv_thread_methods[] = {
   {"equal", luv_thread_equal},
   {"join", luv_thread_join},
+#if LUV_UV_VERSION_GEQ(1, 45, 0)
+  {"getaffinity", luv_thread_getaffinity},
+  {"setaffinity", luv_thread_setaffinity},
+#endif
   {NULL, NULL}
 };
 
